# -*- coding: utf-8 -*-

"""
* Name:         tests/constraints
* Description:  Tests suite for `cognitivefactory.interactive_clustering.constraints`.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""
