# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering
* Description:  Constrained clustering module of the Interactive Clustering package.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""
