# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering
* Description:  Python package used to apply NLP interactive clustering methods.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

from typing import List

__all__: List[str] = []  # noqa: WPS410 (only __variable__ we use)
